/*
 *  libmodem library, a modem control facility.
 *
 *  line_manage.c - modem line control
 *
 *  Copyright (C) 1994,1995,1996,1997  Riccardo Facchetti
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  You can contact the author at this e-mail address:
 *
 *  fizban@tin.it
 */

/*
 * Some termios stuff is stolen from Gert Doering's mgetty-0.21 code.
 * Here is his (C) for mgetty-0.21 package:
 * --------------------------------------------------------------------------
 * The mgetty+sendfax package is Copyright (c) 1993 Gert Doering. You are
 * permitted to do nearly anything you want with this program - redistribute
 * it, use parts of the code in your own programs, ..., but you have to give
 * me credit - do not remove my name.
 * 
 * If you make money by selling mgetty, I want a share.
 * 
 * For more details, see the chapter "copying conditions" in mgetty.texi
 * --------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <setjmp.h>
#include <syslog.h>
#include <ctype.h>
/*
 * Use POSIX termios
 */
#include <termios.h>

#ifdef linux
#include <linux/fs.h>
#include <linux/major.h>
#endif

#include <dial/modems.h>
#include <dial/mdmerrno.h>

#include "line_managePriv.h"
#include "termios_cc.h"

#define BUFSIZE	1024

char *__line_manage = "line_manage.c (C) Riccardo Facchetti, version " LIBMODEM_VER;

static jmp_buf not_opened;
#if 1
struct termios savetty;
#endif

/*
 * Device opened when dial return a valid fd
 */
struct modems *mdmopendevice;
/*
 * If set, cycle all the modems when an error is encountered
 * This don't work if you use the bldial() call
 */
int mdm_cycle = 0;

char *mdm_reply[] = {
"OK",
"ERROR",
"CONNECT",
"NO CARRIER",
"BUSY",
"NO DIALTONE",
"RING",
"NO ANSWER",
NULL
};

static void alarm_jump(int sig) {
    longjmp(not_opened, 1);
}

int dial(char *phonenum)  {
    struct modems *mdl;
    int ret = 0;

    open_mdm_log();
    mdmerrno = 0;
    setmdms();

    while ((mdl = getnextmdm()) != NULL) {
/*
 * if locked, skip to the next modem
 */
        if (mdm_chklock(mdl) == SUCCESS)
		continue;

/*
 * Try to dial...
 */
        ret = bldial(phonenum, mdl->line, mdl->bd);

/*
 * If the user ask it, cycle all the modems on error
 */
	if ((ret < 0) && mdm_cycle) {
/*
 * If cycle is chosen, the meaning of mdm_cycle is not only an on/off, it is
 * the max number of times we cycle. At 0, end of the story: no more retries.
 */
		--mdm_cycle;
/*
 * Log the error and continue (we force the error log because this modem can
 * be in a fault state while the sysadmin don't know the faulty state).
 */
		mdm_log(LOG_INFO,
			"Error encountered while opening %s (%s)\n",
			mdl->line,
			mdmstrerror(ret));

		switch (ret) {
			case -EUNMDMERR:
			case -ENOMDMFILE:
			case -EMDMBUSY:
/*
 * If config file don't exist or unknown error is returned,  or line is busy
 * don't cycle the modems.
 */
				break;
			default:
		    		continue;
		}
	}
/*
 * If the modem is alredy used, skip to the next modem else return the error
 */
	if (ret != EMDMOPEN)
		return ret;
    }

    return (mdmerrno = -ENOMDMDEV);
}

int ldial(char *phonenum, char *line) {
    struct modems *mdl;

    open_mdm_log();
    mdmerrno = 0;

    if ((mdl = getmdmnam(line)) == NULL)
        return mdmerrno;

/*
 * dial at the maximum baud rate
 */
    return bldial(phonenum, line, mdl->bd);
}

int bdial(char *phonenum, int baud)  {
    struct modems *mdl;
    int ret = 0;

    open_mdm_log();
    mdmerrno = 0;
    setmdms();

    while ((mdl = getnextmdm()) != NULL) {
        if (baud <= mdl->bd) {
/*
 * if locked, skip this modem
 */
            if (mdm_chklock(mdl) == SUCCESS)
	    	continue;
/*
 * Try to dial...
 */
            ret = bldial(phonenum, mdl->line, baud);
/*
 * If the user ask it, cycle all the modems on error
 */
	    if ((ret < 0) && mdm_cycle) {
/*
 * If cycle is chosen, the meaning of mdm_cycle is not only an on/off, it is
 * the max number of times we cycle. At 0, end of the story: no more retries.
 */
		--mdm_cycle;
/*
 * Log the error and continue
 */
			mdm_log(LOG_INFO,
				"Error encountered while opening %s (%s)\n",
				mdl->line,
				mdmstrerror(ret));

		    	switch (ret) {
				case -EUNMDMERR:
				case -ENOMDMFILE:
				case -EMDMBUSY:
/*
 * If config file don't exist or unknown error is returned, or line is busy
 * don't cycle the modems.
 */
					break;
				default:
		    			continue;
		    	}
	    }
/*
 * If the modem is alredy used, skip to the next modem else return the error
 */
            if (ret != EMDMOPEN)
                    return ret;
        }
    }

    return (mdmerrno = -ENOMDMDEV);
}

int bldial(char *phonenum, char *line, int baud) {
    struct modems *mdl;
    int fd, ret = 0;

    open_mdm_log();
    mdmerrno = 0;

    if ((mdl = getmdmnam(line)) == NULL)
        return mdmerrno;

    if (mdm_chklock(mdl) == SUCCESS)
	    return (mdmerrno = -EMDMLOCK);
/*
 * lock and open the modem line and set it up for communications
 * remember to unlock the mdl device when returning an error.
 */
    if ((fd = open_modem_setup_line(mdl, baud)) < 0)
        return fd;

/*
 * Now we must save the actual device to have it in hands when hangup
 */
    mdmopendevice = mdl;

/*
 * set up the modem: init string et al
 */
    if ((ret = initialize_modem(fd, mdl)) < 0) {
        mdm_unlock(mdl);
        hangup(fd);
        return ret;
    }

/*
 * finally dial to phonenum
 */
    if ((ret = dial_to_num(fd, mdl, phonenum)) < 0) {
        mdm_unlock(mdl);
        hangup(fd);
        return ret;
    }

/*
 * dial sequence terminated with success, returning fd
 */
    close_mdm_log();
    return fd;
}

static int dial_to_num(int fd, struct modems *device, char *num) {
	char *buffer;
	int ret = -ENOMDMERR;

	buffer = mdmalloc(BUFSIZE);

	if (!buffer)
		return -EMDMEM;
/*
 * DIAL string
 */
    sprintf(buffer, "%s%s%s%s%s%s%c",
		    device->cp,
		    device->ds,
		    device->dp,
		    num,
                    device->de,
		    device->ce,
		    MODEM_RETURN);
    ret = talk_to_modem(fd, device, buffer, 1);

    mdmfree(buffer);

    return ret;
}

static int initialize_modem(int fd, struct modems *device) {
	char *buffer;
	int ret;

	buffer = mdmalloc(BUFSIZE);

	if (!buffer)
		return -EMDMEM;
/*
 * INIT string
 */
	sprintf(buffer,
		"%s%s%s%c", device->cp, device->im, device->ce, MODEM_RETURN);
	ret = talk_to_modem(fd, device, buffer, 0);

	mdmfree(buffer);

	return ret;
}

static int open_modem_setup_line(struct modems *device, int baud) {
	struct termios trm;
	int fd = -1;
	struct stat st;
	void (*old_alarm)(int);
	int old_alarm_value = 0;
	char *buffer;

	buffer = mdmalloc(BUFSIZE);

	if (!buffer)
		return -EMDMEM;

#if defined(MDM_CHECK_DEVICES)
reopen_line:
#endif

	if (mdm_lock(device) == FAILURE) {
		mdmfree(buffer);
		return (mdmerrno = -EMDMLOCK);
	}

	if (setjmp(not_opened) == 0) {
/*
 * build the modem device filename
 */
		sprintf(buffer, "/dev/%s", device->line);

/*
 * set up the alarm for jump if timeout
 */
		old_alarm_value = alarm(0);
		old_alarm = signal(SIGALRM, alarm_jump);
		alarm(2);

		fd = open(buffer, O_RDWR|O_NOCTTY|O_NDELAY);
		if (fd >= 0) {
			int n = 0;
/*
 * Cancel the O_NDELAY flag.
 */
#ifdef SBAGLIATO
			n = fcntl(fd, F_GETFL, 0);
#else
			n = fcntl(fd, F_GETFL);
#endif
			n = fcntl(fd, F_SETFL, n & ~O_NDELAY);
		}

	}
/*
 * Restore the old alarm handler and scheduled timeout
 */
	alarm(0);
	signal(SIGALRM, old_alarm);
	alarm(old_alarm_value);
	if (fd < 0) {
		mdmfree(buffer);
		return (mdmerrno = -EMDMOPEN);
	}

/*
 * fstat it to get informations
 */
	if (fstat(fd, &st) == -1) {
		mdm_unlock(device);
		close(fd);
		mdmfree(buffer);
		return (mdmerrno = -EMDMSTAT);
	}
/*
 * Now check if it is a character device.
 */
	if (!S_ISCHR(st.st_mode)) {
		mdm_unlock(device);
		close(fd);
		mdmfree(buffer);
		return (mdmerrno = -EMDMNOCHR);
	}

#if defined(MDM_CHECK_DEVICES)
/*
 * Check if we are using a callout device. If not, close the ttyS device and
 * change it with the cua one.
 */

/*
 * Now check the minor device (is better to check the major one after minor)
 *
 * Warning: Linux specific.
 */
#define MDM_MINOR_START		64
#define MDM_MINOR_END		127

	if ((MINOR(st.st_rdev) < MDM_MINOR_START) ||
	    (MINOR(st.st_rdev) > MDM_MINOR_END)) {
		mdm_unlock(device);
		close(fd);
		mdmfree(buffer);
		return (mdmerrno = -EMDMBADMIN);
	}

/*
 * Now check if major is okay:
 * major == TTYAUX_MAJOR (callout device). If == TTY_MAJOR, reopen callout one.
 */
	 if ((MAJOR(st.st_rdev) != TTY_MAJOR) &&
	     (MAJOR(st.st_rdev) != TTYAUX_MAJOR)) {
		mdm_unlock(device);
		close(fd);
		mdmfree(buffer);
		return (mdmerrno = -EMDMBADMAJ);
	}

/*
 * Now check if major == TTY_MAJOR, if so, close, change device and reopen.
 */
	if (MAJOR(st.st_rdev) == TTY_MAJOR) {
		mdm_unlock(device);
		close(fd);
/*
 * Now we are going to change the device from ttyS one to cua one.
 * The problem is that we are going to lock /dev/cua?? so if another person
 * will attempt to open the same /dev/ttyS?? line, we will have a timeout
 * error on open, and only on the second (right) open, (open of /dev/cua??)
 * we will find the lock of the line.
 */
		sprintf(device->line, "cua%d", MINOR(st.st_rdev));
		goto reopen_line;
	}
#endif /* MDM_CHECK_DEVICES */
/*
 * Set up the line for working with modem (POSIX termios)
 * Retry 60 times (1 sec sleep) while tcgetattr succedes.
 */
	{
		int retr=0, result=0;

		while (retr < 60) {
			if (tcgetattr(fd, &trm) != -1) {
				result = 1;
				break;
			}
			sleep(1);
			retr++;
		}
		if (result == 0) {
			mdm_unlock(device);
			close(fd);
#if defined(DEBUG)
			mdm_log(LOG_DEBUG,
				"tcgetattr retries: too many, dropping line\n");
#endif /* DEBUG */
			mdmfree(buffer);
			return (mdmerrno = -EMDMGETA);
		}
#if defined(DEBUG)
		mdm_log(LOG_DEBUG, "tcgetattr retries: %d\n", retr);
#endif /* DEBUG */
	}

#if 1
/*
 * save tty state
 */
	memcpy(&savetty, &trm, sizeof(struct termios));
#endif

/*
 * raw mode
 */
	trm.c_iflag &= ~(IGNBRK | IGNCR | INLCR | ICRNL | IUCLC |
			 IXANY | IXON | IXOFF | INPCK | ISTRIP);
	trm.c_iflag = BRKINT|IGNPAR|IXON|IXANY;
	trm.c_oflag = OPOST|TAB3;
	trm.c_lflag = ~(ICANON | ISIG | ECHO | ECHONL | ECHOE | ECHOK);
	trm.c_cflag &= ~(CSIZE|CSTOPB|PARENB|PARODD
#if 0
/*
 * Do touch use CLOCAL (we are calling out).
 */
						|CLOCAL);
#else
						);
#endif
	trm.c_cflag |= CS8|CREAD|HUPCL;
	trm.c_lflag = ECHOK|ECHOE|ECHO|ISIG|ICANON;
	trm.c_cc[VEOF] = 0x04;
#if defined(VEOL) && VEOL < TIONCC
	trm.c_cc[VEOL] = 0;
#endif
#if defined(VSWTCH)
	trm.c_cc[VSWTCH] = 0;
#endif

/*
 * set speed
 */
	if (device->bd < baud) {
		mdm_unlock(device);
		close(fd);
		mdmfree(buffer);
		return (mdmerrno = -EMDMBAUD);
	}
	cfsetospeed(&trm, baud);
	cfsetispeed(&trm, baud);

/*
 * set default c_cc flags
 */
	trm.c_cc[VQUIT]  = CQUIT;
	trm.c_cc[VKILL]  = CKILL;
	trm.c_cc[VEOF]   = CEOF;
#if defined(VEOL) && VEOL < TIONCC
	trm.c_cc[VEOL] = CEOL;
#endif
#if defined(VSTART) && VSTART < TIONCC
	trm.c_cc[VSTART] = CSTART;
#endif
#if defined(VSTOP) && VSTOP < TIONCC
	trm.c_cc[VSTOP] = CSTOP;
#endif
#if defined(VSUSP) && VSUSP < TIONCC
	trm.c_cc[VSUSP] = CSUSP;
#endif
#if defined(VSWTCH) && VSWTCH < TIONCC
	trm.c_cc[VSWTCH] = CSWTCH;
#endif
    /* the following are for SVR4.2 (and higher) */
#if defined(VDSUSP) && VDSUSP < TIONCC
	trm.c_cc[VDSUSP] = CDSUSP;
#endif
#if defined(VREPRINT) && VREPRINT < TIONCC
	trm.c_cc[VREPRINT] = CRPRNT;
#endif
#if defined(VDISCARD) && VDISCARD < TIONCC
	trm.c_cc[VDISCARD] = CFLUSH;
#endif
#if defined(VWERASE) && VWERASE < TIONCC
	trm.c_cc[VWERASE] = CWERASE;
#endif
#if defined(VLNEXT) && VLNEXT < TIONCC
	trm.c_cc[VLNEXT] = CLNEXT;
#endif

/*
 * set mode raw
 */
	trm.c_iflag &= ( IXON | IXOFF | IXANY );    /* clear all flags except */
                        /* xon / xoff handshake */
	trm.c_oflag  = 0;                /* no output processing */
	trm.c_lflag  = 0;                /* no signals, no echo */
	trm.c_cc[VMIN]  = 1;                /* disable line buffering */
	trm.c_cc[VTIME] = 0;

/*
 * set flow control
 */
	trm.c_cflag &= ~CRTSCTS;
	trm.c_iflag &= ~(IXON|IXOFF|IXANY);
	if ( device->fc == FLOW_HARD )
		trm.c_cflag |= CRTSCTS;
	if ( device->fc == FLOW_XON_XOFF )
		trm.c_iflag |= IXOFF|IXON|IXANY;

/*
 * set the new attributes, flush output, discard all not read input
 */
	if (tcsetattr(fd, TCSAFLUSH, &trm) == -1) {
		mdm_unlock(device);
		close(fd);
		mdmfree(buffer);
		return (mdmerrno = -EMDMSETA);
	}

	mdmfree(buffer);

	return fd;
}

static int get_command(char *buffer, char *cmd) {
	char *nbuffer;
	char *p;

	nbuffer = mdmalloc(BUFSIZE);

	if (!nbuffer)
		return -EMDMEM;

	if ((p = strstr(buffer, "\r\n"))) {
		/* zero \r */
		*p = '\0';
		/* jump to next token */
		p += 2;
		strcpy(nbuffer, p);
		strcpy(cmd, buffer);
		strcpy(buffer, nbuffer);
		mdmfree(nbuffer);
		return SUCCESS;
	}
	mdmfree(nbuffer);
	return FAILURE;
}

static
int talk_to_modem(int fd, struct modems *device, char *command, int isdial) {
	fd_set readfds;
	struct timeval tmout;
	int ret;
	int cmd_found = FAILURE;
	int retries = SYNC_RTRY;
	int mdm_read_retries = READ_RTRY;
	char *cmdbuffer;
	char *bigbuffer;
	char *cmdmodem;
	char *buffer;

	cmdbuffer = mdmalloc(BUFSIZE);

	if (!cmdbuffer)
		return -EMDMEM;

	bigbuffer = mdmalloc(BUFSIZE);

	if (!bigbuffer) {
		mdmfree(cmdbuffer);
		return -EMDMEM;
	}

	cmdmodem = mdmalloc(BUFSIZE);

	if (!cmdmodem) {
		mdmfree(cmdbuffer);
		mdmfree(bigbuffer);
		return -EMDMEM;
	}

	buffer = mdmalloc(BUFSIZE);

	if (!buffer) {
		mdmfree(cmdbuffer);
		mdmfree(bigbuffer);
		mdmfree(cmdmodem);
		return -EMDMEM;
	}

#if defined(DEBUG)
	mdm_log(LOG_DEBUG, "writing >> %s << to modem\n", command);
#endif /* DEBUG */

/*
 * write the command to modem
 */

	if (write(fd, command, strlen(command)) != strlen(command)) {
		mdmfree(cmdbuffer);
		mdmfree(bigbuffer);
		mdmfree(cmdmodem);
		mdmfree(buffer);
		return (mdmerrno = -EMDMWRITE);
	}

	mdm_log(LOG_INFO, "written >> %s << to modem\n", command);

    /*
     * I need to check this thing.
     */
	sleep (1);

	while (!cmd_found) {

		memset(bigbuffer, 0, BUFSIZE);

		FD_ZERO(&readfds);
		FD_SET(fd, &readfds);
		tmout.tv_sec = device->dl;
		tmout.tv_usec = 0;
 /* vediamo se uno sleep mi permette di sincronizzarmi */
		sleep(1);

#if 1
		ret = select(FD_SETSIZE, &readfds, NULL, NULL, &tmout);
#else
	/*
	 * Just 1 FD to wait on (read from fd).
	 */
		ret = select(1, &readfds, NULL, NULL, &tmout);
#endif

		if (ret < 0) {
			mdmfree(cmdbuffer);
			mdmfree(bigbuffer);
			mdmfree(cmdmodem);
			mdmfree(buffer);
			return (mdmerrno = -EMDMSEL);
		}

		if (!FD_ISSET(fd, &readfds)) {
			mdmfree(cmdbuffer);
			mdmfree(bigbuffer);
			mdmfree(cmdmodem);
			mdmfree(buffer);
			return (mdmerrno = -EMDMNOFDSET);
		}
        
		if ((ret = read(fd, bigbuffer, BUFSIZE)) < 0) {
			mdmfree(cmdbuffer);
			mdmfree(bigbuffer);
			mdmfree(cmdmodem);
			mdmfree(buffer);
			return (mdmerrno = -EMDMREAD);
		}
        
		bigbuffer[ret] = '\0';

#if defined(DEBUG)
		mdm_log(LOG_DEBUG, "read >> %s << from modem\n", bigbuffer);
#endif /* DEBUG */

		strcat(cmdbuffer, bigbuffer);

		if (get_command(cmdbuffer, cmdmodem) == FAILURE) {
			if (!(mdm_read_retries--)) {
				mdmfree(cmdbuffer);
				mdmfree(bigbuffer);
				mdmfree(cmdmodem);
				mdmfree(buffer);
				return (mdmerrno = -EMDMRDRT);
			}
			continue;
		}

/*
 * Reset read retries: we have found something.
 */
		mdm_read_retries = READ_RTRY;

		if (!strcmp(command, cmdmodem)) {
/*
 * local echo enabled!!! strip out the command from modem answer!!!!
 * that is "command\r\nOK\r\n" or something like that
 */
			if (get_command(cmdbuffer, cmdmodem) == FAILURE) {
				if (!(mdm_read_retries--)) {
					mdmfree(cmdbuffer);
					mdmfree(bigbuffer);
					mdmfree(cmdmodem);
					mdmfree(buffer);
					return (mdmerrno = -EMDMRDRT);
				}
				continue;
			}
		mdm_read_retries = READ_RTRY;
		}

		mdm_log(LOG_INFO, "checking >> %s << from modem\n", cmdmodem);

		if (check_for_string(cmdmodem, buffer) == SUCCESS)
			cmd_found = 1;
		if (!(retries--)) {
			mdmfree(cmdbuffer);
			mdmfree(bigbuffer);
			mdmfree(cmdmodem);
			mdmfree(buffer);
			return (mdmerrno = -EMDMSYNC);
		}
	}

/*
 * If modem return ERROR, we must return the error now
 * if it is RINGing we must return.
 */
	if (!strcmp(buffer, "ERROR")) {
		mdmfree(cmdbuffer);
		mdmfree(bigbuffer);
		mdmfree(cmdmodem);
		mdmfree(buffer);
		return (mdmerrno = -EMDMERROR);
	}

	if (!strcmp(buffer, "RING")) {
		mdmfree(cmdbuffer);
		mdmfree(bigbuffer);
		mdmfree(cmdmodem);
		mdmfree(buffer);
		return (mdmerrno = -EMDMRING);
	}

	if (isdial) {
/*
 * check the answer against 'CONNECT' or BUSY
 */
		if (!strcmp(buffer, "CONNECT")) {
			mdmfree(cmdbuffer);
			mdmfree(bigbuffer);
			mdmfree(cmdmodem);
			mdmfree(buffer);
			return (mdmerrno = -ENOMDMERR);
		}
		if (!strcmp(buffer, "BUSY")) {
			mdmfree(cmdbuffer);
			mdmfree(bigbuffer);
			mdmfree(cmdmodem);
			mdmfree(buffer);
			return (mdmerrno = -EMDMBUSY);
		}
		if (!strcmp(buffer, "NO DIALTONE")) {
			mdmfree(cmdbuffer);
			mdmfree(bigbuffer);
			mdmfree(cmdmodem);
			mdmfree(buffer);
			return (mdmerrno = -EMDMTONE);
		}
		if (!strcmp(buffer, "NO ANSWER")) {
			mdmfree(cmdbuffer);
			mdmfree(bigbuffer);
			mdmfree(cmdmodem);
			mdmfree(buffer);
			return (mdmerrno = -EMDMANSW);
		}
	} else {
/*
 * check the answer against 'OK' ? No err : Unk err
 */
		if (!strcmp(buffer, "OK")) {
			mdmfree(cmdbuffer);
			mdmfree(bigbuffer);
			mdmfree(cmdmodem);
			mdmfree(buffer);
			return (mdmerrno = -ENOMDMERR);
		}
	}
	mdmfree(cmdbuffer);
	mdmfree(bigbuffer);
	mdmfree(cmdmodem);
	mdmfree(buffer);
	return (mdmerrno = -EUNMDMERR);
}

static int check_for_string(char *from, char *to) {
	int i;

	for (i=0; mdm_reply[i]; i++)
		if (!strncasecmp(from, mdm_reply[i], strlen(mdm_reply[i]))) {
			strcpy(to, mdm_reply[i]);
			return SUCCESS;
		}
	return FAILURE;
}

int close_nohangup(int fd) {
	struct termios trm;

	tcgetattr(fd, &trm);
	trm.c_cflag &= ~HUPCL;
	tcsetattr(fd, TCSANOW, &trm);

	close(fd);
/*
 * fd closed, but modem still on-line
 */
	return (mdmerrno = -ENOMDMERR);
}

int hangup(int fd) {
	struct termios trm;
	char *buffer;
	int n;

	if (fd < 0)
		return (mdmerrno = -EMDMBADFD);

	if (mdmopendevice == NULL)
		return hangup_nodev(fd);
/*
 * hangup with termios (speed to B0). just to be sure :)
 */
	if (tcgetattr(fd, &trm) != -1) {
		cfsetospeed(&trm, B0);
		cfsetispeed(&trm, B0);
/*
 * hangup NOW lowering DTR (0 speed) and sleep 2 seconds to allow modem
 * to hangup
 */
		tcsetattr(fd, TCSANOW, &trm);
		sleep(2);
	}

#if 1
/*
 * reset line stat
 */
	tcsetattr(fd, TCSANOW, &savetty);
#endif

/*
 * Set the O_NDELAY flag.
 */
	n = fcntl(fd, F_GETFL, 0);
	n = fcntl(fd, F_SETFL, n | O_NDELAY);
/*
 * ATTENTION+HANGUP string with nodelay, because we don't want block here
 * and sleep for allow modem to hangup
 */
	buffer = malloc(BUFSIZE);

	if (!buffer)
		return -EMDMEM;

	sprintf(buffer, "%s%s%s%s%c", mdmopendevice->at,  mdmopendevice->cp,
				      mdmopendevice->hu, mdmopendevice->ce,
				      MODEM_RETURN);
	write (fd, buffer, strlen(buffer));
	mdmfree(buffer);
	sleep(2);

/*
 * Cancel the O_NDELAY flag.
 */
	n = fcntl(fd, F_GETFL, 0);
	n = fcntl(fd, F_SETFL, n & ~O_NDELAY);

/*
 * close and unlock
 */
	close(fd);
	mdm_unlock(mdmopendevice);

	mdmopendevice = NULL;

	if (mdmerrno < 0)
		return mdmerrno;
	return (mdmerrno = -ENOMDMERR);
}

int hangup_nodev(int fd) {
	struct termios trm;

/*
 * hangup with termios (speed to B0). just to be sure :)
 */
	if (tcgetattr(fd, &trm) != -1) {
		cfsetospeed(&trm, B0);
		cfsetispeed(&trm, B0);
/*
 * hangup NOW lowering DTR (0 speed) and sleep 2 seconds to allow modem
 * to hangup
 */
		tcsetattr(fd, TCSANOW, &trm);
		sleep(2);
	}

#if 1
/*
 * reset line stat
 */
	tcsetattr(fd, TCSANOW, &savetty);
#endif

/*
 * close and unlock
 */
	close(fd);

	if (mdmerrno < 0)
		return mdmerrno;
	return (mdmerrno = -ENOMDMERR);
}

struct modems *getopenmodem(void) {
	static struct modems openmdm;

	if (mdmopendevice != NULL) {
		memcpy(&openmdm, mdmopendevice, sizeof(struct modems));
		return &openmdm;
	}
	else
		return NULL;
}

int clear_line(int fd, int waittime) {
	char buffer[2];
	int bytes;

	struct termios tty, savetty;

	if (tcgetattr(fd, &tty) == -1) {
#if defined(DEBUG)
		mdm_log(LOG_DEBUG, "tcgetattr: cannot clear line\n");
#endif /* DEBUG */
		return (mdmerrno = -EMDMGETA);
	}
    
	savetty = tty;

	tty.c_lflag &= ~ICANON;
	tty.c_cc[VMIN] = 0;
	tty.c_cc[VTIME] = waittime;

	tcsetattr(fd, TCSANOW, &tty);

	while ( read( fd, buffer, 1 ) > 0 && bytes < 10000 )
		bytes++;

/* reset terminal settings */
	tcsetattr(fd, TCSANOW, &savetty);
    
	return SUCCESS;
}
